/*
 * Copyright (c) 1995 John Kohl.  All rights reserved.
 *
 * This software contains fragments derived from code which is 
 * Copyright (c) 1995 Rolf Grossmann.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by John Kohl and
 *	Rolf Grossmann.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 *	$Id: mixer_main.c,v 1.6 1995/07/09 23:12:30 jtk Exp $
 */

#include "forms.h"
#include "mixer.h"
#include "mixer_c.h"

int debug;

#define DPRINTF(x) if (debug) printf x

int
get_encoding(const char *encstring)
{
    if (strcmp(encstring, AudioElinear) == 0)
	return AUDIO_ENCODING_LINEAR;
    if (strcmp(encstring, AudioEmulaw) == 0)
	return AUDIO_ENCODING_ULAW;
    if (strcmp(encstring, AudioEalaw) == 0)
	return AUDIO_ENCODING_ALAW;
    if (strcmp(encstring, AudioEpcm16) == 0)
	return AUDIO_ENCODING_PCM16;
    if (strcmp(encstring, AudioEpcm8) == 0)
	return AUDIO_ENCODING_PCM8;
    if (strcmp(encstring, AudioEadpcm) == 0)
	return AUDIO_ENCODING_ADPCM;
    warnx("encoding `%s' not recognized", encstring);
    return -1;
}

const char *
encoding_text(int val)
{
    switch (val) {
/* PCM16 is the same as linear
   case AUDIO_ENCODING_LINEAR: */
    default:
	return  AudioElinear;
    case AUDIO_ENCODING_ULAW:
	return  AudioEmulaw;
    case AUDIO_ENCODING_ALAW:
	return  AudioEalaw;
    case AUDIO_ENCODING_PCM16:
	return  AudioEpcm16;
    case AUDIO_ENCODING_PCM8:
	return  AudioEpcm8;
    case AUDIO_ENCODING_ADPCM:
	return  AudioEadpcm;
    }
}

mixer_ctrl_t masterinfo;
int masterinfo_found = 0, recording_found = 0;
int mutes_found = 0;
int input_class;
mixer_ctrl_t playinfo;
mixer_ctrl_t recinfo;
mixer_ctrl_t muteinfo;

void
get_devinfo(int mixer, DB **namedb, DB **indexdb)
{
    DB *dev_db, *indx_db;
    DBT key, datum;
    mixer_devinfo_t di, storedi;
    int error;
    devdata_db_t dbitem;
    int first;

    dev_db = dbopen(NULL, O_RDWR|O_CREAT, 0666, DB_HASH, NULL);
    if (dev_db == NULL)
	err(1, "open name database");
    
    indx_db = dbopen(NULL, O_RDWR|O_CREAT, 0666, DB_HASH, NULL);
    if (indx_db == NULL)
	err(1, "open devindex database");
    
    di.index = 0;
    for (di.index = 0;
	 ioctl(mixer, AUDIO_MIXER_DEVINFO, &di) == 0;
	 di.index++, di.label.name[0] = '\0') {
	dbitem.index = di.index;
	strcpy(dbitem.name, di.label.name);
	dbitem.type = di.type;
	dbitem.devclass = di.mixer_class;
	dbitem.previndex = di.prev;
	/* mark master volume specially, so we don't put it in the
	   list of things eligible for generic handling */
	if (strcmp(di.label.name, AudioNvolume) == 0) {
	    DPRINTF(("found masterinfo index %d\n", di.index));
	    masterinfo_found = 1;
	    masterinfo.un.value.num_channels = 2;
	    masterinfo.dev = di.index;
	    masterinfo.type = AUDIO_MIXER_VALUE;
	    if (ioctl(mixer, AUDIO_MIXER_READ, &masterinfo)) {
		masterinfo.un.value.num_channels = 1;
		if (ioctl(mixer, AUDIO_MIXER_READ, &masterinfo))
		    warn("can't get master info");
	    }
	    dbitem.srcflags = MIXER_MASTER_VOL;
	} else if (strcmp(di.label.name, AudioNrecord) == 0) {
	    DPRINTF(("found recording gain %d\n", di.index));
	    recording_found = 1;
	    recinfo.un.value.num_channels = 2;
	    recinfo.dev = di.index;
	    recinfo.type = AUDIO_MIXER_VALUE;
	    if (ioctl(mixer, AUDIO_MIXER_READ, &recinfo)) {
		recinfo.un.value.num_channels = 1;
		if (ioctl(mixer, AUDIO_MIXER_READ, &recinfo))
		    warn("can't get recording info");
	    }
	    dbitem.srcflags = MIXER_RECORD_VOL;
	} else if (strcmp(di.label.name, AudioNspeaker) == 0) {
	    /* speaker can't be manipulated */
	    dbitem.srcflags = MIXER_MASTER_VOL;
	} else if (strcmp(di.label.name, AudioCInputs) == 0) {
	    DPRINTF(("inputs class index %d class %d\n", di.index,
		     di.mixer_class));
	    input_class = di.mixer_class;
	} else if (strcmp(di.label.name, AudioNmute) == 0 &&
		   di.type == AUDIO_MIXER_ENUM) {
	    DPRINTF(("mute at index %d: prev is %d\n", di.index, di.prev));
	    /* we need to alter the name to be unique so that we don't
	       get hash collisions */
	    (void) sprintf(di.label.name, AudioNmute "%d", di.index);
	    mutes_found++;
	} else if (first && di.type == AUDIO_MIXER_VALUE) {
	    char mixtxt[MAX_AUDIO_DEV_LEN + 16];
	    playinfo.un.value.num_channels = 2;
	    playinfo.dev = dbitem.index;
	    playinfo.type = AUDIO_MIXER_VALUE;
	    if (ioctl(mixer_fd, AUDIO_MIXER_READ, &playinfo)) {
		playinfo.un.value.num_channels = 1;
		if (ioctl(mixer, AUDIO_MIXER_READ, &playinfo))
		    warn("can't fetch mixer level for default `%s'",
			 dbitem.name);
	    }
	    sprintf(mixtxt, "%s volume", dbitem.name);
	    fl_set_object_label(fd_mixerform->srcvol_text, mixtxt);
	    sprintf(mixtxt, "%s mode", dbitem.name);
	    fl_set_object_label(fd_mixerform->srcmix_text, mixtxt);
	    dbitem.srcflags = 0;
	    first = 0;
	} else
	    dbitem.srcflags = 0;

	datum.data = &dbitem;
	datum.size = sizeof(dbitem);
	key.data = di.label.name;
	key.size = strlen(di.label.name)+1; /* include the NUL */

	DPRINTF(("storing key `%s' (%d) dev %d\n",
		 key.data, key.size, dbitem.index));
	if ((*dev_db->put)(dev_db, &key, &datum, 0) != 0)
	    err(1, "build database");
	key.data = &di.index;
	key.size = sizeof(di.index);

	DPRINTF(("storing key %d name `%s'\n",
		 *(int *)key.data, dbitem.name));
	if ((*indx_db->put)(indx_db, &key, &datum, 0) != 0)
	    err(1, "build database");

	/* XXX Assume the source list comes after all the elements. */
	if (strcmp(di.label.name, AudioNsource) == 0 &&
	    di.type == AUDIO_MIXER_ENUM) {
	    register int i;
	    DPRINTF(("found sources: %d\n", di.index));
	    for (i = 0; i < di.un.e.num_mem; i++) {
		key.data = &di.un.e.member[i].ord;
		key.size = sizeof(di.un.e.member[i].ord);
		if ((*indx_db->get)(indx_db, &key, &datum, 0) == 0) {
		    bcopy(datum.data, &dbitem, sizeof(dbitem));
		    dbitem.srcflags |= MIXER_REC_SRC;
		    datum.data = &dbitem;
		    datum.size = sizeof(dbitem);
		    if ((*indx_db->put)(indx_db, &key, &datum, 0) != 0)
			err(1, "build database");
		    key.data = dbitem.name;
		    key.size = strlen(dbitem.name)+1;
		    if ((*dev_db->put)(dev_db, &key, &datum, 0) != 0)
			err(1, "build database");
		    DPRINTF(("marked `%s' as recordable\n", dbitem.name));
		} else
		    warnx("entry ordering problem for `%s' in source list",
			  di.un.e.member[i].label.name);
	    }
	}
    }

    *namedb = dev_db;
    *indexdb = indx_db;
}

void
set_sliders(FL_OBJECT *slider_l,
	    FL_OBJECT *slider_r,
	    FL_OBJECT *slider_both,
	    FL_OBJECT *m_button,
	    FL_OBJECT *st_button,
	    mixer_ctrl_t *mixinfo)
{
    fl_freeze_form(fd_mixerform->mixerform);
    if (mixinfo->un.value.num_channels == 2) {
	fl_show_object(slider_l);
	fl_show_object(slider_r);
	if (slider_both)
	    fl_hide_object(slider_both);
	fl_set_slider_value(slider_l,
			    mixinfo->un.value.level[AUDIO_MIXER_LEVEL_LEFT]);
	fl_set_slider_value(slider_r,
			    mixinfo->un.value.level[AUDIO_MIXER_LEVEL_RIGHT]);
	fl_set_button(st_button,1);
	fl_set_button(m_button,0);
    } else {
	if (slider_both) {
	    fl_hide_object(slider_l);
	    fl_hide_object(slider_r);
	    fl_show_object(slider_both);
	}
	fl_set_slider_value(slider_l,
			    mixinfo->un.value.level[AUDIO_MIXER_LEVEL_MONO]);
	fl_set_slider_value(slider_r,
			    mixinfo->un.value.level[AUDIO_MIXER_LEVEL_MONO]);
	if (slider_both)
	    fl_set_slider_value(slider_both,
				mixinfo->un.value.level[AUDIO_MIXER_LEVEL_MONO]);
	fl_set_button(st_button,0);
	fl_set_button(m_button,1);
    }
    fl_unfreeze_form(fd_mixerform->mixerform);
}

DB *dev_db_name, *dev_db_index;

FD_mixerform *fd_mixerform;
int mixer_fd = -1;
int snd_fd = -1;
int *encarray, *encbackarray, *mutearray;

int snd_count = 0;

void
open_snd(int mustwork)
{
    if (snd_count) {
	snd_count++;
	return;
    }
    snd_fd = open("/dev/sound", O_RDWR, 0); /* XXX define _PATH_SOUND ? */
    if (snd_fd < 0) {
	if (mustwork)
	    err(1, "Can't open sound device");
	else
	    warn("Can't open sound device");
    }
    snd_count++;
    return;
}

void
close_snd()
{
    if (--snd_count)
	return;
    (void) close(snd_fd);
    snd_fd = -1;
}

void
usage(const char *progname)
{
    err(1, "%s: Usage: %s [-d] { xforms arguments }\n", progname, progname);
}

int
main(int argc, char *argv[])
{
    DBT key, datum;
    mixer_ctrl_t ctl;
    int i, j, k, v, error;
    char *p;
    extern int optind;
    struct audio_info ainfo, ascratch;
    audio_encoding_t encinfo;

    char inputval[32];
    devdata_db_t dbitem;

    mixer_fd = open("/dev/mixer", O_RDONLY, 0); /* XXX define _PATH_MIXER ? */
    if (mixer_fd < 0)
	err(1, "Can't open mixer device");

    open_snd(1);
    if (ioctl(snd_fd, AUDIO_GETINFO, &ainfo))
	err(1, "can't get current audio info");


    fl_initialize(argv[0], 0, 0, 0, &argc, argv);
    while ((i = getopt(argc, argv, "d")) != -1) {
	switch (i) {
	case 'd':
	    debug = 1;
	    break;
	default:
	    usage(argv[0]);
	}
    }

    fd_mixerform = create_form_mixerform();

    get_devinfo(mixer_fd, &dev_db_name, &dev_db_index);


    fl_hide_object(fd_mixerform->rec_l); /* XXX notyet */
    fl_hide_object(fd_mixerform->rec_r); /* XXX notyet */
    if (masterinfo_found) {
	fl_set_slider_bounds(fd_mixerform->master_l,255.0,0.0);
	fl_set_slider_return(fd_mixerform->master_l,0);
	fl_set_slider_bounds(fd_mixerform->master_r,255.0,0.0);
	fl_set_slider_return(fd_mixerform->master_r,0);
	fl_set_slider_step(fd_mixerform->master_l,1.0);
	fl_set_slider_step(fd_mixerform->master_r,1.0);
	fl_set_slider_precision(fd_mixerform->master_l,0);
	fl_set_slider_precision(fd_mixerform->master_r,0);
	set_sliders(fd_mixerform->master_l, fd_mixerform->master_r, NULL,
		    fd_mixerform->master_mono, fd_mixerform->master_stereo,
		    &masterinfo);
    } else  {
	fl_hide_object(fd_mixerform->master_l);
	fl_hide_object(fd_mixerform->master_r);
	fl_hide_object(fd_mixerform->master_mono);
	fl_hide_object(fd_mixerform->master_stereo);
	fl_hide_object(fd_mixerform->mvol_label);
	fl_hide_object(fd_mixerform->omode_label);
    }	
    if (mutes_found) {
	/* set up array for mutes; filled in below */
	mutearray = (int *)malloc(mutes_found * sizeof(int));
    }

    fl_set_slider_bounds(fd_mixerform->rec_both,255.0,0.0);
    fl_set_slider_return(fd_mixerform->rec_both,0);
/*    fl_set_slider_bounds(fd_mixerform->rec_r,255.0,0.0);*/
    fl_set_slider_bounds(fd_mixerform->play_l,255.0,0.0);
    fl_set_slider_return(fd_mixerform->play_l,0);
    fl_set_slider_bounds(fd_mixerform->play_r,255.0,0.0);
    fl_set_slider_return(fd_mixerform->play_r,0);
    fl_set_slider_bounds(fd_mixerform->play_both,255.0,0.0);
    fl_set_slider_return(fd_mixerform->play_both,0);

    fl_set_slider_step(fd_mixerform->rec_both,1.0);
/*    fl_set_slider_step(fd_mixerform->rec_r,1.0);*/
    fl_set_slider_step(fd_mixerform->play_l,1.0);
    fl_set_slider_step(fd_mixerform->play_r,1.0);
    fl_set_slider_step(fd_mixerform->play_both,1.0);

    fl_set_slider_precision(fd_mixerform->rec_both,0);
/*    fl_set_slider_precision(fd_mixerform->rec_r,0);*/
    fl_set_slider_precision(fd_mixerform->play_l,0);
    fl_set_slider_precision(fd_mixerform->play_r,0);
    fl_set_slider_precision(fd_mixerform->play_both,0);


    if (recording_found)
	fl_set_slider_value(fd_mixerform->rec_both,
           (double) recinfo.un.value.level[AUDIO_MIXER_LEVEL_MONO]);
    else
	fl_set_slider_value(fd_mixerform->rec_both, 255.0); /* XXX */

/*    fl_set_slider_value(fd_mixerform->rec_r,255.0);*/

    set_sliders(fd_mixerform->play_l, fd_mixerform->play_r,
		fd_mixerform->play_both,
		fd_mixerform->play_mono, fd_mixerform->play_stereo,
		&playinfo);

    if (ainfo.record.channels == 2)
	fl_set_button(fd_mixerform->rec_stereo,1);
    else
	fl_set_button(fd_mixerform->rec_mono,1);

    /* XXX do anything about full duplex stuff? AUDIO_GETFD/SETFD ? */

    /*
     * XXX must match bit order  in <sys/audioio.h>
     */
    fl_addto_choice(fd_mixerform->dac_mode,"Off");
    fl_addto_choice(fd_mixerform->dac_mode,"Play");
    fl_addto_choice(fd_mixerform->dac_mode,"Record");
    fl_addto_choice(fd_mixerform->dac_mode,"Play & Record");
    if (ioctl(snd_fd, AUDIO_GETFD, &i) == 0 && i == 0)
	fl_set_choice_item_mode(fd_mixerform->dac_mode, 4, FL_PUP_GRAY);

    /* choices are 1-origin */
    fl_set_choice(fd_mixerform->dac_mode, ainfo.mode + 1);
    DPRINTF(("initial duplexmode %d\n", ainfo.mode));

    /* check out each encoding before putting into the list */

    DPRINTF(("initial encodings: play %s record %s\n",
	     encoding_text(ainfo.play.encoding),
	     encoding_text(ainfo.record.encoding)));

    /*
     * count 'em up
     */
    for (i = 0, j = 0, encinfo.index = i;
	 ioctl(snd_fd, AUDIO_GETENC, &encinfo) == 0;
	 encinfo.index = ++i)
	if (j < encinfo.format_id)
	    j = encinfo.format_id;

    if (i) {
	encarray = (int *)malloc(i * sizeof(int));
	encbackarray = (int *)malloc((j+1) * sizeof(int));
	if (encarray == 0 || encbackarray == 0)
	    err(1, "can't allocate memory for encoding array");
	while (j >= 0)
	    encbackarray[j--] = 1;	/* init all to point at first name */
	/* copy them in */
	for (i = 0, encinfo.index = i;
	     ioctl(snd_fd, AUDIO_GETENC, &encinfo) == 0;
	     encinfo.index = ++i) {
	    encarray[i] = encinfo.format_id;
	    /* establish reverse map for those that are usable.
	       choice index is 1-based */
	    encbackarray[encinfo.format_id] = i + 1;
	    fl_addto_choice(fd_mixerform->rec_encoding,
			    encinfo.name);
	    fl_addto_choice(fd_mixerform->play_encoding,
			    encinfo.name);
	    if (encinfo.format_id == ainfo.record.encoding)
		fl_set_choice(fd_mixerform->rec_encoding, i + 1);
	    if (encinfo.format_id == ainfo.play.encoding)
		fl_set_choice(fd_mixerform->play_encoding, i + 1);
	}
	
    }


    sprintf(inputval, "%d", ainfo.play.sample_rate);
    fl_set_input(fd_mixerform->playrate, inputval);
    sprintf(inputval, "%d", ainfo.record.sample_rate);
    fl_set_input(fd_mixerform->recrate, inputval);

    sprintf(inputval, "%d", ainfo.play.precision);
    fl_set_input(fd_mixerform->playprec, inputval);
    sprintf(inputval, "%d", ainfo.record.precision);
    fl_set_input(fd_mixerform->recprec, inputval);

    /* set up the recording, mixer source lists */
    for (i = 1, j = 1, k = 0,
	     error = (*dev_db_name->seq)(dev_db_name, &key, &datum, R_FIRST);
	 error == 0;
	 error = (*dev_db_name->seq)(dev_db_name, &key, &datum, R_NEXT)) {
	bcopy(datum.data, &dbitem, sizeof(dbitem));
	DPRINTF(("`%s'(%d) is index %d\n", key.data, key.size, i));
	if (dbitem.type == AUDIO_MIXER_VALUE &&
	    dbitem.srcflags & MIXER_REC_SRC) {
	    /* eligible as recording source */
	    fl_addto_choice(fd_mixerform->rec_source, dbitem.name);
	    if (dbitem.index == ainfo.record.port)
		fl_set_choice(fd_mixerform->rec_source, i);
	    i++;
	}
	if (dbitem.type == AUDIO_MIXER_VALUE &&
	    (dbitem.srcflags & MIXER_MASTER_VOL) == 0 &&
	    dbitem.devclass == input_class) {
	    /* eligible for generic volume control */
	    fl_addto_choice(fd_mixerform->mix_source, dbitem.name);
	    if (dbitem.index == playinfo.dev)
		fl_set_choice(fd_mixerform->mix_source, j);
	    j++;
	}
	if (mutes_found && dbitem.type == AUDIO_MIXER_ENUM &&
	    memcmp(dbitem.name, AudioNmute, strlen(AudioNmute)) == 0 &&
	    dbitem.devclass == input_class) {
	    devdata_db_t ndbitem;
	     /* It's a mute.  Fetch the name it mutes from its previndex */
	     key.data = &dbitem.previndex;
	     key.size = sizeof(dbitem.previndex);
	     if ((*dev_db_index->get)(dev_db_index, &key, &datum, 0) == 0) {
		 bcopy(datum.data, &ndbitem, sizeof(dbitem));
		 fl_addto_choice(fd_mixerform->mute_choice, ndbitem.name);
		 mutearray[k++] = dbitem.index;
	     }
	}
    }
    if (k > 0) {
	fl_set_choice(fd_mixerform->mute_choice, 1);
	muteinfo.dev = mutearray[0];
	muteinfo.type = AUDIO_MIXER_ENUM;
	if (ioctl(mixer_fd, AUDIO_MIXER_READ, &muteinfo) == 0)
	    /* muteinfo.un.ord set means device is muted. */
	    fl_set_button(fd_mixerform->mute_btn, muteinfo.un.ord ? 1 : 0);
    } else {
	fl_hide_object(fd_mixerform->mute_choice);
	fl_hide_object(fd_mixerform->mute_btn);
    }	
    close_snd();

    /* show the first form */
    fl_show_form(fd_mixerform->mixerform,FL_PLACE_CENTER,FL_FULLBORDER,"mixerform");
    while (fl_do_forms())
	;
    return 0;
}
